--[[ ---------------------------------------------------------------------------

Acheron_Graph: graph reports for Acheron

Acheron_Graph is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Acheron_Graph is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with Acheron_Graph.	If not, see <http://www.gnu.org/licenses/>.

----------------------------------------------------------------------------- ]]

--[[ ---------------------------------------------------------------------------
entry Memento :
entry.eventType :
	"START" > cast start
	"FAILED" > cast fail
	"SUCCESS" > cast success, to detect instacast spells
entry.startTime		-- Only for START events
entry.endTime		-- Only for START events
entry.timeStamp		
entry.spellId 		-- Only for START/FAILED/SUCCESS, DAMAGE events may have if relevant 
entry.spellName		-- same as spellId
entry.srcName
entry.srcGUID
entry.targetsGUID	-- for START events contains a table of [dstGUID] = [AmountHealed or true if not a heal event]
entry.pushBacks		-- number of pushback events, only used for START events that have been delayed by pushback
entry.failType		-- contains fail type if cast has failed, only used for START events
entry.isSuccess		-- for START events for which cast time has been completed
-- For entries copied by value on unit death only
entry.deathUnitAmountHealed	-- for START events, amount of heal done on dead unit if applicable (only defined when trying to display)
----------------------------------------------------------------------------- ]]

--[[ ---------------------------------------------------------------------------
	 Addon and libraries
----------------------------------------------------------------------------- ]]


Acheron_Graph = LibStub("AceAddon-3.0"):NewAddon("Acheron_Graph", "AceConsole-3.0", "AceHook-3.0", "AceEvent-3.0", "AceTimer-3.0")
local Acheron = Acheron
local Acheron_Graph = Acheron_Graph

local L = LibStub("AceLocale-3.0"):GetLocale("Acheron_Graph")

--[[ ---------------------------------------------------------------------------
	 Ace3 initialization
----------------------------------------------------------------------------- ]]
function Acheron_Graph:OnInitialize()
    -- Initialize persistent addon variables
    self.combatSpellLogs = {}
    self.timeOffset = 0
    -- Create frame
    self:CreateFrame()
end

function Acheron_Graph:OnEnable()
    -- Hooking various Acheron methods
    self:Hook(Acheron, "ClearReports", "hook_ClearReports", true)
    self:SecureHook(Acheron, "PopulateEntries", "hook_PopulateEntries", true)
    
    --[[ Registering events and hooking stuff dealing with interacters spell log
    -- Hooking various Acheron methods
    self:SecureHook(Acheron, "TrackEvent", "hook_TrackEvent", true)
    self:SecureHook(Acheron, "COMBAT_LOG_EVENT_UNFILTERED", "hook_COMBAT_LOG_EVENT_UNFILTERED", true)
    -- Registering tick for update of offset timer
	self:ScheduleRepeatingTimer("registerCombatLogEvent", 10)
    -- Registering events
    self:RegisterEvent("UNIT_SPELLCAST_START", "handleSpellcastStart")
    self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED", "handleSpellcastStart")
    self:RegisterEvent("UNIT_SPELLCAST_DELAYED", "handleDelayEvent")
    self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_UPDATE", "handleDelayEvent")
    self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_START", "handleSpellcastStart") ]]--
    
end

function Acheron_Graph:OnDisable()
    -- Unhooking Acheron methods
end

--[[ ---------------------------------------------------------------------------
	 Hooks
----------------------------------------------------------------------------- ]]
function Acheron_Graph:hook_COMBAT_LOG_EVENT_UNFILTERED(self, event, timeStamp, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)--XXX
	local prefix, suffix, special = strsplit("_", eventType)
	
	local log = Acheron_Graph.combatSpellLogs[srcGUID] 
	if log then
		-- For spell events concerning a watched unit attempts to fill in the corresponding spellcast log
		if prefix == "SPELL" then
			local i = log.last
			-- cycling backward through log (limit at 4 sec) to attempt finding relevant spellcast entry
			while i>=log.first and log[i].timeStamp >= timeStamp - 10 do
				local spellId, spellName = ...
				if log[i].eventType == "START" and log[i].spellName == spellName then
					--if spellName == "Penance" then self:Print("Penance : ", eventType, " on ", dstGUID) end
					log[i].spellId = spellId
					if not log[i].targetsGUID[dstGUID] then
						log[i].targetsGUID[dstGUID] = true
					end
					-- For HEAL events retrieve amount healed and add it
					if suffix == "HEAL" or special == "HEAL" then
						local amountHealed = select(4, ...)
						-- Add the amount healed to targetsGUID table
						if log[i].targetsGUID[dstGUID]~=true then
							log[i].targetsGUID[dstGUID] = log[i].targetsGUID[dstGUID] + amountHealed
						else
							log[i].targetsGUID[dstGUID] = amountHealed
						end
					end
					break
				end
				i = i-1
			end
		end
		-- For spell events dealing with FAILED spells, attempt to fill in corresponding spellcast log
		if eventType == "SPELL_CAST_FAILED" then
			local spellId = select(1, ...)
			local spellName = select(2, ...)
			local failType = select(4, ...)
			
			-- handles repercussions on casting spell	
			local i = log.last
			local lastTimeStamp = log[log.last].timeStamp
			-- cycling backward through log to find stopped cast and updating endTime
			while i>=log.first do
				if log[i].eventType == "START"
				and log[i].spellName == spellName			
				and log[i].endTime~=log[i].startTime then 	-- avoids failing instacast
					if log[i].timeStamp >= lastTimeStamp - (log[i].endTime - log[i].startTime)/1000 and not log[i].isSuccess then -- time check
						log[i].spellId = spellId
						log[i].endTime = GetTime() * 1000
						log[i].failType = failType
					end
					break
				end
				i = i-1
			end
		end
	end
	
  	-- Handles damage events
  	if Acheron.combatLogs[dstGUID] and suffix == "DAMAGE" then
  		Acheron_Graph:handleDamageEvent(event, timeStamp, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
  	end
end

function Acheron_Graph:hook_TrackEvent(self, eventType, timeStamp, msg, dstGUID, amount, action, source, spell, isCrit, isCrush, spellId)
	-- Process death event
	if eventType == "DEATH" then
		local latestLog = Acheron.combatLogs[dstGUID].logs[Acheron.combatLogs[dstGUID].last]
		latestLog.interactersSpellLogs = Acheron_Graph:getInteractersSpellLogs(dstGUID)
	end
end

function Acheron_Graph:hook_ClearReports(self, name)
	if not name then
		Acheron_Graph.combatSpellLogs = {}
	end
end

--[[ ---------------------------------------------------------------------------
	 handling time offset computation by periodically registering
	 COMBAT_LOG_EVENT_UNFILTERED
----------------------------------------------------------------------------- ]]
function Acheron_Graph:handleCombatLogEvent(event, timeStamp, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
	-- computes time difference between combat log timestamp and local time
	self.timeOffset = timeStamp - GetTime()
	--self:Print("Offset updated")
	self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
end

function Acheron_Graph:registerCombatLogEvent()
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED", "handleCombatLogEvent")
end

--[[ ---------------------------------------------------------------------------
	 Spell casts events handlers
----------------------------------------------------------------------------- ]]
-- SpellCastStart
function Acheron_Graph:handleSpellcastStart(event, unit, spellName, ...)
	local GUID = UnitGUID(unit)
	if not Acheron.combatLogs[GUID] then return end
	--self:Print(event, spellName)

	-- UNIT_SPELLCAST_START/SUCCEEDED is triggered twice if raid member is in party, skipping
	if strsub(unit, 1, 4) == "raid" then
		if UnitInParty(unit) or UnitIsPlayer(unit) then return end
	end
			
	local entry = {}
	
	local UnitCastingInfo = UnitCastingInfo
	if event == "UNIT_SPELLCAST_CHANNEL_START" then
		UnitCastingInfo = UnitChannelInfo
	end
	local spellNameCast, _, _, _, startTime, endTime = UnitCastingInfo(unit)
	--if spellNameCast == "Penance" then self:Print(startTime, endTime) end
	-- If spell is instant
	if not startTime or not endTime then
		local time = GetTime()*1000
		entry.startTime = time
		entry.endTime = time
		entry.isSuccess = true
	-- Else
	else
		entry.startTime = startTime
		entry.endTime = endTime
		entry.isSuccess = false
	end
	
	entry.pushBacks = 0
	
	-- Before continuing check that there is not a spell with same name and cast time ~= 0 before in spell log, if there is then tag it as success
	if event == "UNIT_SPELLCAST_SUCCEEDED" and self.combatSpellLogs[GUID] then
		local log = self.combatSpellLogs[GUID]
		local i = log.last
		while i>=log.first and log[i].timeStamp >= log[log.last].timeStamp - 5 do
			if log[i].eventType == "START" and log[i].spellName == spellName and log[i].endTime-log[i].startTime ~= 0 then
				log[i].isSuccess = true
				return
			end
			i = i-1
		end
	end
			
	entry.eventType = "START"
	entry.timeStamp = self.timeOffset+GetTime()
	--entry.spellId = ???
	entry.spellName = spellName
	entry.srcName = UnitName(unit)
	entry.srcGUID = GUID
	entry.targetsGUID = {}
	entry.msg = string.format("Spell cast : %s", spellName)
	
	self:AddEntryToLog(GUID, entry)
end

function Acheron_Graph:handleDelayEvent(event, unit, ...)
	local GUID = UnitGUID(unit)
	if not self.combatSpellLogs[GUID] then return end
	-- handles repercussions on casting spell
	local UnitCastingInfo = UnitCastingInfo
	if event == "UNIT_SPELLCAST_CHANNEL_UPDATE" then -- hack to get channeling info
		UnitCastingInfo = UnitChannelInfo
	end	
	local spellName, _, _, _, startTime, endTime = UnitCastingInfo(unit)
	--self:Print(event, spellName)
	if spellName then
		local log = self.combatSpellLogs[GUID]
		local i = log.last
		-- cycling backward through log to find delayed start and updating endTime
		while i>=log.first and log[i].timeStamp >= log[log.last].timeStamp - (endTime-startTime)/1000 do
			if log[i].eventType == "START" and log[i].spellName == spellName then
				log[i].endTime = endTime
				log[i].pushBacks = log[i].pushBacks + 1
				break
			end
			i = i-1
		end
	end
end

function Acheron_Graph:handleDamageEvent(event, timeStamp, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
	local entry = {}
	
	entry.timeStamp = timeStamp
	entry.srcName = srcName
	entry.srcGUID = srcGUID
	entry.targetsGUID = {}
	entry.targetsGUID[dstGUID] = true
	entry.msg = CombatLog_OnEvent(Blizzard_CombatLog_CurrentSettings, timeStamp, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
	entry.eventType = "DAMAGE"
	
	local prefix, suffix, special = strsplit("_", eventType)
	if prefix == "SPELL" then
		entry.spellId = select(1, ...)
		entry.spellName = select(2, ...)
	end
	
	self:AddEntryToLog(dstGUID, entry)
end

function Acheron_Graph:AddEntryToLog(GUID, entry)
	if not self.combatSpellLogs[GUID] then
		self.combatSpellLogs[GUID] = {
			first = 1,
			last = 0,
		}
	end
	
	local log = self.combatSpellLogs[GUID]
	
	-- Add the entry to the spell log
	log.last = log.last + 1
	log[log.last] = entry
	
	-- Remove any entries that are older than our threshold
	for i = log.first, log.last do
		if (log[i].timeStamp >= (entry.timeStamp - Acheron:GetProfileParam("history"))) then
			log.first = i;
			break
		else
			log[i] = nil
		end	
	end
end

--[[ ---------------------------------------------------------------------------
	 Called in case of a death event, process all spell logs to find logs
	 from actors that interacted with dead unit (including itself)
----------------------------------------------------------------------------- ]]
local function copyTable(t)
	local r = {} 
	for k, v in pairs(t) do
		if type(v) == "table" then
			r[k] = copyTable(v)
		else
			r[k] = v
		end
	end
	return r
end

function Acheron_Graph:getInteractersSpellLogs(deathGUID)
	local interactersSpellLog = {}
	for spellCasterGUID, combatSpellLog in pairs(self.combatSpellLogs) do
		for i = combatSpellLog.first, combatSpellLog.last do
			local entry = combatSpellLog[i]
			-- If spellcaster is found to interact with dead unit (or is the dead unit) copy by value everything in log that is relevant
			if entry.targetsGUID[deathGUID] or spellCasterGUID == deathGUID then
				local savedLog = {}
				savedLog.last = combatSpellLog.last
				savedLog.first = combatSpellLog.first
				for j = combatSpellLog.first, combatSpellLog.last do
					savedLog[j] = copyTable(combatSpellLog[j])
				end
				-- flag the log as pertaining to dead unit if necessary
				if spellCasterGUID == deathGUID then
					savedLog.isDead = true
				end
				interactersSpellLog[spellCasterGUID] = savedLog
				break
			end
		end
	end
	
	return interactersSpellLog
end
